/*
$Id: r5test.c,v 1.8 2008/05/18 21:28:33 eric Exp eric $
Copyright 2007, 2008 Eric Smith <eric@brouhaha.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License version 3 as
published by the Free Software Foundation.  Note that I am not
granting permission to redistribute or modify this program under the
terms of any later version of the General Public License.

This program is distributed in the hope that it will be useful, but
WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#define _LARGEFILE64_SOURCE

#include <fcntl.h>
#include <inttypes.h>
#include <stdarg.h>
#include <stdbool.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>

#include "mt19937ar.h"


char *progname;

bool show_progress;


char magic[8] = "-r5test-";


int64_t arg_block_size = 512;
int64_t arg_stripe_size = 128;
int64_t arg_count = 1000;
int64_t arg_skip_count = 0;


void usage (FILE *f)
{
  fprintf (f, "%s <function> <options>... <file>...\n", progname);
  fprintf (f, "functions:\n");
  fprintf (f, "    extract <src>... <dest>   extract RAID 5 data into a file\n");
  fprintf (f, "    wr5t <dest>...            write test data striped in RAID 5 fashion\n");
  fprintf (f, "    rd5t <src>...             read test data striped in RAID 5 fashion\n");
  fprintf (f, "    wt <dest>                 write test data\n");
  fprintf (f, "    rt <src>                  read test data\n");
  fprintf (f, "options:\n");
  fprintf (f, "    -bs <n>                   block size (bytes), default %" PRId64 "\n", arg_block_size);
  fprintf (f, "    -ss <n>                   stripe size (blocks), default %" PRId64" d\n", arg_stripe_size);
  fprintf (f, "    -c <n>                    count (blocks or stripes), default %" PRId64 "\n", arg_count);
  fprintf (f, "                              (-1 on read to continue to end of file)\n");
  fprintf (f, "    -sc <n>                   skip count (stripes), default %" PRId64 "\n", arg_skip_count);
}


void fatal (int ret, char *fmt, ...) __attribute__ ((noreturn));

void fatal (int ret, char *fmt, ...)
{
  va_list ap;

  if (fmt)
    {
      fprintf (stderr, "fatal error: ");
      va_start (ap, fmt);
      vfprintf (stderr, fmt, ap);
      va_end (ap);
    }
  if (ret == 1)
    usage (stderr);
  exit (ret);
}


void *alloc (size_t size)
{
  void *p = malloc (size);
  if (! p)
    fatal (2, "failed to allocate memory\n");
  return p;
}


// init by calling with s=0, ignore result
bool seconds_elapsed (int s)
{
  static struct timeval t1 = { 0, 0 };
  struct timeval t2;
  bool status = false;
  time_t sd;

  if (s == 0)
    {
      t1.tv_sec = 0;
      t1.tv_usec = 0;
      return true;
    }

  gettimeofday (& t2, NULL);

  sd = (t2.tv_sec - t1.tv_sec);
  if (t2.tv_usec < t1.tv_usec)
    sd--;
  if (sd > s)
    status = true;
  //printf ("%d %d\n", sd, status);
  if (status)
    memcpy (& t1, & t2, sizeof (struct timeval));

  return status;
}


void put_uint32_le (uint8_t *buffer, uint32_t value)
{
  size_t i;

  for (i = 0; i < sizeof (value); i++)
    {
      *(buffer++) = value & 0xff;
      value >>= 8;
    }
}


void put_uint64_le (uint8_t *buffer, uint64_t value)
{
  size_t i;

  for (i = 0; i < sizeof (value); i++)
    {
      *(buffer++) = value & 0xff;
      value >>= 8;
    }
}


void gen_block_content (size_t block_size,
			uint64_t block_index,
			uint8_t *buffer)
{
  if (block_size < (sizeof (magic) + sizeof (block_index)))
    fatal (2, "block size too small\n");

  // magic
  memcpy (buffer, magic, sizeof (magic));
  buffer += sizeof (magic);
  block_size -= sizeof (magic);

  // block index, written in little-endian order
  put_uint64_le (buffer, block_index);
  buffer += sizeof (block_index);
  block_size -= sizeof (block_index);

  init_genrand ((block_index >> 32) ^ block_index);

  // fill rest of buffer with pseudo-random data based on block index
  while (block_size >= sizeof (uint32_t))
    {
      put_uint32_le (buffer, genrand_int32 ());
      buffer += sizeof (uint32_t);
      block_size -= sizeof (uint32_t);
    }

  // pad any remaining bytes with zeros
  if (block_size)
    memset (buffer, 0, block_size);
}


bool block_write (int fd,
		  uint8_t *buffer,
		  size_t count)
{
  while (count)
    {
      ssize_t r = write (fd, buffer, count);
      if (r < 0)
	return false;  // could be EOF or error
      buffer += r;
      count -= r;
    }
  return true;
}


bool write_test_blocks (int fd,
			size_t block_size,
			uint64_t start_index,
			int64_t block_count_req)
{
  bool status = true;  // OK
  uint64_t block_index = start_index;
  int64_t block_count = 0;
  uint8_t *buffer;

  buffer = alloc (block_size);

  (void) seconds_elapsed (0);

  while ((block_count_req < 0) || (block_count < block_count_req))
    {
      if (show_progress && seconds_elapsed (1))
	{
	  printf ("\rwriting block %ld", block_count);
	  fflush (stdout);
	}

      gen_block_content (block_size, block_index, buffer);
      status = block_write (fd, buffer, block_size);
      if (! status)
	{
	  // should check for EOF or error
	  break;
	}
      block_index++;
      block_count++;
    }

  if (show_progress)
    printf ("\n");

  if (! status)
    perror ("error writing block");

  printf ("wrote %ld blocks\n", block_count);

  free (buffer);
  return status;
}


bool write_test (char *fn, 
		 size_t block_size,
		 uint64_t start_index,
		 int64_t block_count_req)
{
  bool status;
  int fd;

  fd = open (fn, O_WRONLY | O_CREAT, 0666);
  if (fd < 0)
    {
      perror ("open for write");
      fatal (2, "can't open output file\n");
    }

  status = write_test_blocks (fd, block_size, start_index, block_count_req);

  close (fd);

  return status;
}


bool block_read (int fd,
		 uint8_t *buffer,
		 size_t count)
{
  while (count)
    {
      ssize_t r = read (fd, buffer, count);
      if (r < 0)
	{
	  return false;  // error
	}
      if (r == 0)
	{
	  return false;  // eof
	}
      buffer += r;
      count -= r;
    }
  return true;
}


bool read_test_blocks (int fd,
		       size_t block_size,
		       uint64_t start_index,
		       int64_t block_count_req)
{
  bool status = true;  // OK
  uint64_t block_index = start_index;
  int64_t block_count = 0;
  int64_t mismatch_count = 0;
  uint8_t *buffer, *buffer2;

  buffer = alloc (block_size);
  buffer2 = alloc (block_size);

  (void) seconds_elapsed (0);

  while ((block_count_req < 0) || (block_count < block_count_req))
    {
      if (show_progress && seconds_elapsed (1))
	{
	  printf ("\rreading block %ld", block_count);
	  fflush (stdout);
	}

      gen_block_content (block_size, block_index, buffer);

      status = block_read (fd, buffer2, block_size);
      if (! status)
	{
	  // should check for EOF or error
	  break;
	}

      if (memcmp (buffer, buffer2, block_size))
	{
	  if (show_progress)
	    printf ("\n");
	  printf ("block %ld mismatch\n", block_count);
	  mismatch_count++;
	}

      block_index++;
      block_count++;
    }

  if (show_progress)
    printf ("\n");

  if (! status)
    perror ("error reading block");

  printf ("read %ld blocks\n", block_count);
  printf ("%ld mismatches\n", mismatch_count);

  free (buffer);
  return status;
}


bool read_test (char *fn, 
		size_t block_size,
		uint64_t start_index,
		int64_t block_count_req)
{
  bool status;
  int fd;

  fd = open (fn, O_RDONLY);
  if (fd < 0)
    {
      perror ("open for read");
      fatal (2, "can't open input file\n");
    }

  status = read_test_blocks (fd, block_size, start_index, block_count_req);

  close (fd);

  return status;
}


void dump_hex (uint8_t *b, size_t block_size)
{
  while (block_size--)
    printf (" %02x", *(b++));
  printf ("\n");
}


void block_xor (uint8_t *dest, uint8_t *src, size_t block_size)
{
  while (block_size--)
    *(dest++) ^= *(src++);
}


void gen_r5_stripe_content (int file_count,
			    int parity_index,
			    size_t block_size,
			    int blocks_per_stripe,
			    int start_block_index,
			    uint8_t *buffer)
{
  int i, j;
  uint64_t block_index = start_block_index;
  uint8_t *b;
  uint8_t *pb;

  b = buffer;

  pb = buffer + parity_index * block_size * blocks_per_stripe;
  memset (pb, 0, block_size * blocks_per_stripe);

  for (i = 0; i < file_count; i++)
    {
      if (i == parity_index)
	continue;
      b = buffer + i * block_size * blocks_per_stripe;
      pb = buffer + parity_index * block_size * blocks_per_stripe;
      for (j = 0; j < blocks_per_stripe; j++)
	{
	  gen_block_content (block_size, block_index++, b);
	  block_xor (pb, b, block_size);
	  b += block_size;
	  pb += block_size;
	}
    }
}


bool write_r5_test_stripes (int disk_count,
			    int *fd,
			    size_t block_size,
			    int blocks_per_stripe,
			    int64_t stripe_count_req)
{
  bool status = true;  // OK
  int64_t stripe_count = 0;
  uint8_t *buf;
  int parity_index = 0;

  buf = alloc (disk_count * block_size * blocks_per_stripe);

  (void) seconds_elapsed (0);

  while ((stripe_count_req < 0) || (stripe_count < stripe_count_req))
    {
      int64_t start_block_index = stripe_count * blocks_per_stripe * (disk_count - 1);
      int i;
      uint8_t *dp;

      if (show_progress && seconds_elapsed (1))
	{
	  printf ("\rwriting stripe %ld (block %ld)",
		  stripe_count,
		  start_block_index);
	  fflush (stdout);
	}

      gen_r5_stripe_content (disk_count,
			     parity_index,
			     block_size,
			     blocks_per_stripe,
			     start_block_index,
			     buf);

      dp = buf;
      for (i = 0; i < disk_count; i++)
	{
	  status = block_write (fd [i], dp, block_size * blocks_per_stripe);
	  if (! status)
	    {
	      // should check for EOF or error
	      goto fail;
	    }
	    dp += block_size * blocks_per_stripe;
	}

      stripe_count++;

      parity_index = (parity_index + 1) % disk_count;
    }

  if (show_progress)
    printf ("\n");

  if (! status)
    perror ("error writing block");

  printf ("wrote %ld stripes\n", stripe_count);

 fail:
  free (buf);
  return status;
}


bool write_r5_test (int disk_count,
		    char *fn [], 
		    size_t block_size,
		    int blocks_per_stripe,
		    int64_t stripe_count_req)
{
  bool status;
  int i;
  int *fd;

  fd = alloc (disk_count * sizeof (int));

  for (i = 0; i < disk_count; i++)
    {
      fd [i] = open (fn [i], O_WRONLY | O_CREAT, 0666);
      if (fd [i] < 0)
	{
	  perror ("open for write");
	  fatal (2, "can't open output file '%s'\n", fn [i]);
	}
    }

  status = write_r5_test_stripes (disk_count,
				  fd,
				  block_size,
				  blocks_per_stripe,
				  stripe_count_req);

  for (i = 0; i < disk_count; i++)
    close (fd [i]);

  return status;
}


bool read_r5_test_stripes (int disk_count,
			   int *fd,
			   size_t block_size,
			   int blocks_per_stripe,
			   int64_t stripe_count_req)
{
  bool status = true;  // OK
  int64_t stripe_count = 0;
  int64_t mismatch_count = 0;
  uint8_t *disk_buf;   // disk stripes
  uint8_t *exp_buf;    // expected data
  int parity_index = 0;

  disk_buf = alloc (disk_count * block_size * blocks_per_stripe);
  exp_buf  = alloc (disk_count * block_size * blocks_per_stripe);

  (void) seconds_elapsed (0);

  while ((stripe_count_req < 0) || (stripe_count < stripe_count_req))
    {
      int64_t start_block_index = stripe_count * blocks_per_stripe * (disk_count - 1);
      int i;
      uint8_t *dp;

      if (show_progress && seconds_elapsed (1))
	{
	  printf ("\rreading stripe %ld (block %ld)",
		  stripe_count,
		  start_block_index);
	  fflush (stdout);
	}

      gen_r5_stripe_content (disk_count,
			     parity_index,
			     block_size,
			     blocks_per_stripe,
			     start_block_index,
			     exp_buf);

      dp = disk_buf;
      for (i = 0; i < disk_count; i++)
	{
	  status = block_read (fd [i], dp, block_size * blocks_per_stripe);
	  if (! status)
	    {
	      // should check for EOF or error
	      goto fail;
	    }
	    dp += block_size * blocks_per_stripe;
	}

      if (memcmp (exp_buf, disk_buf, disk_count * block_size * blocks_per_stripe))
	{
	  if (show_progress)
	    printf ("\n");
	  printf ("stripe %ld mismatch\n", stripe_count);
	  mismatch_count++;
	}

      stripe_count++;

      parity_index = (parity_index + 1) % disk_count;
    }

  if (show_progress)
    printf ("\n");

  if (! status)
    perror ("error reading block");

  printf ("read %ld stripes\n", stripe_count);
  printf ("%ld mismatches\n", mismatch_count);

 fail:
  free (disk_buf);
  free (exp_buf);
  return status;
}


bool read_r5_test (int disk_count,
		   char *fn [], 
		   size_t block_size,
		   int blocks_per_stripe,
		   int64_t stripe_count_req)
{
  bool status;
  int i;
  int *fd;

  fd = alloc (disk_count * sizeof (int));

  for (i = 0; i < disk_count; i++)
    {
      fd [i] = open (fn [i], O_RDONLY);
      if (fd [i] < 0)
	{
	  perror ("open for read");
	  fatal (2, "can't open input file '%s'\n", fn [i]);
	}
    }

  status = read_r5_test_stripes (disk_count,
				 fd,
				 block_size,
				 blocks_per_stripe,
				 stripe_count_req);

  for (i = 0; i < disk_count; i++)
    close (fd [i]);

  return status;
}


// this could be a lot more efficient
bool r5_check_parity (int src_disk_count,
		      int stripe_size,      // bytes
		      uint8_t *buf)
{
  int i, j;
  for (i = 0; i < stripe_size; i++)
    {
      uint8_t pv = 0;
      for (j = 0; j < src_disk_count; j++)
	{
	  pv ^= buf [j * stripe_size + i];
	}
      if (pv != 0)
	return false;
    }
  return true;
}


bool r5_extract_stripes (int src_disk_count,
			 int *src_fd,
			 int dest_fd,
			 size_t block_size,
			 int blocks_per_stripe,
			 int64_t first_stripe,
			 int64_t stripe_count_req)
{
  bool status = true;  // OK
  int64_t stripe_count = first_stripe;
  int64_t parity_error_count = 0;
  uint8_t *buf;   // disk stripes

  buf = alloc (src_disk_count * block_size * blocks_per_stripe);

  (void) seconds_elapsed (0);

  while ((stripe_count_req < 0) || (stripe_count < stripe_count_req))
    {
      int64_t start_block_index = stripe_count * blocks_per_stripe * (src_disk_count - 1);
      int parity_index = stripe_count % src_disk_count;
      int i;
      uint8_t *dp;

      if (show_progress && seconds_elapsed (1))
	{
	  printf ("\rreading stripe %ld (block %ld)",
		  stripe_count,
		  start_block_index);
	  fflush (stdout);
	}

      dp = buf;
      for (i = 0; i < src_disk_count; i++)
	{
	  status = block_read (src_fd [i], dp, block_size * blocks_per_stripe);
	  if (! status)
	    {
	      // should check for EOF or error
	      goto fail;
	    }
	    dp += block_size * blocks_per_stripe;
	}

      status = r5_check_parity (src_disk_count,
				block_size * blocks_per_stripe,
				buf);
      if (! status)
	{
	  if (show_progress)
	    printf ("\n");
	  printf ("stripe %ld parity error\n", stripe_count);
	  parity_error_count++;
	}

      dp = buf;
      for (i = 0; i < src_disk_count; i++)
	{
	  if (i != parity_index)
	    {
	      status = block_write (dest_fd, dp, block_size * blocks_per_stripe);
	      if (! status)
		{
		  printf ("stripe %ld write error\n", stripe_count);
		  goto fail;
		}
	    }
	  dp += block_size * blocks_per_stripe;
	}

      stripe_count++;
    }

  if (show_progress)
    printf ("\n");

  if (! status)
    perror ("error reading block");

  printf ("read %ld stripes\n", stripe_count);
  printf ("%ld parity errors\n", parity_error_count);

 fail:
  free (buf);
  return status;
}


bool r5_extract (int src_disk_count,
		 char *src_fn [], 
		 char *dest_fn,
		 size_t block_size,
		 int blocks_per_stripe,
		 int64_t first_stripe,
		 int64_t stripe_count_req)
{
  bool status;
  int i;
  int *src_fd;
  int dest_fd;

  off64_t src_offset = first_stripe * block_size * blocks_per_stripe;
  off64_t dest_offset = first_stripe * block_size * blocks_per_stripe *(src_disk_count - 1);

  printf ("first_stripe: %ld\n", first_stripe);
  printf ("src_offset: %ld\n", src_offset);
  printf ("dest_offset: %ld\n", dest_offset);

  src_fd = alloc (src_disk_count * sizeof (int));

  for (i = 0; i < src_disk_count; i++)
    {
      src_fd [i] = open (src_fn [i], O_RDONLY);
      if (src_fd [i] < 0)
	{
	  perror ("open for read");
	  fatal (2, "can't open input file '%s'\n", src_fn [i]);
	}
      if (first_stripe != 0)
	{
	  if (lseek64 (src_fd [i], src_offset, SEEK_SET) < 0)
	    {
	      perror ("seek input file");
	      fatal (2, "can't seek input file '%s'\n", src_fn [i]);
	    }
	}
    }

  dest_fd = open (dest_fn, O_WRONLY | O_CREAT, 0666);
  if (dest_fd < 0)
    {
      perror ("open for write");
      fatal (2, "can't open output file '%s'\n", dest_fn);
    }
  if (first_stripe != 0)
    {
      if (lseek64 (dest_fd, dest_offset, SEEK_SET) < 0)
	{
	  perror ("seek output file");
	  fatal (2, "can't seek output file '%s'\n", src_fn [i]);
	}
    }

  status = r5_extract_stripes (src_disk_count,
			       src_fd,
			       dest_fd,
			       block_size,
			       blocks_per_stripe,
			       first_stripe,
			       stripe_count_req);

  for (i = 0; i < src_disk_count; i++)
    close (src_fd [i]);

  close (dest_fd);

  return status;
}


int main (int argc, char *argv [])
{
  int start_index = 0;

  char *cmd;

  progname = argv [0];
  show_progress = true;

  if (argc < 3)
    fatal (1, NULL);

  cmd = argv [1];

  argv += 2;
  argc -= 2;

  while (argc && (argv [0][0] == '-'))
    {
      if (strcmp (argv [0], "-bs") == 0)
	{
	  if (argc < 2)
	    fatal (1, "argument needed for %s option\n", argv [0]);
	  arg_block_size = atoll (argv [1]);
	}
      else if (strcmp (argv [0], "-ss") == 0)
	{
	  if (argc < 2)
	    fatal (1, "argument needed for %s option\n", argv [0]);
	  arg_stripe_size = atoll (argv [1]);
	}
      else if (strcmp (argv [0], "-c") == 0)
	{
	  if (argc < 2)
	    fatal (1, "argument needed for %s option\n", argv [0]);
	  arg_count = atoll (argv [1]);
	}
      else if (strcmp (argv [0], "-sc") == 0)
	{
	  if (argc < 2)
	    fatal (1, "argument needed for %s option\n", argv [0]);
	  arg_skip_count = atoll (argv [1]);
	}
      else
	fatal (1, "unrecognized option '%s'\n", argv [0]);
      argv += 2;
      argc -= 2;
    }

  printf ("block size = %" PRId64 "\n", arg_block_size);
  printf ("stripe size = %" PRId64 "\n", arg_stripe_size);
  printf ("count = %" PRId64 "\n", arg_count);
  printf ("skip count = %" PRId64 "\n", arg_skip_count);

  if (strcmp (cmd, "wt") == 0)
    {
      write_test (argv [0],           // fn
		  arg_block_size,     // block_size
		  start_index,        // start_index
		  arg_count);         // block_count_req
    }
  else if (strcmp (cmd, "rt") == 0)
    {
      read_test (argv [0],            // fn
		 arg_block_size,      // block_size
		 start_index,         // start_index
		 arg_count);          // block_count_req
    }
  else if (strcmp (cmd, "wr5t") == 0)
    {
      write_r5_test (argc,  // disk count
		     argv,            // fn array
		     arg_block_size,  // block size
		     arg_stripe_size, // blocks per stripe
		     arg_count);      // stripe_count_req
    }
  else if (strcmp (cmd, "rr5t") == 0)
    {
      read_r5_test (argc,             // disk count
		    argv,             // fn array
		    arg_block_size,   // block size
		    arg_stripe_size,  // blocks per stripe
		    arg_count);       // stripe_count_req
    }
  else if (strcmp (cmd, "extract") == 0)
    {
      r5_extract (argc - 1,           // source disk count
		  argv,               // source fn array
		  argv [argc - 1],    // dest disk fn
		  arg_block_size,     // block size
		  arg_stripe_size,    // blocks per stripe
		  arg_skip_count,     // first stripe
		  arg_count);         // stripe_count_req
    }
  else
    fatal (1, NULL);

  exit (0);
}
